from plone.distribution.testing import INTEGRATION_TESTING
from pytest_plone import fixtures_factory


pytest_plugins = ["pytest_plone"]


globals().update(fixtures_factory(((INTEGRATION_TESTING, "integration"),)))



from AccessControl import getSecurityManager
from plone.distribution.api import distribution as dist_api
from plone.distribution.api import site as site_api
from plone.distribution.utils.request import extract_browser_language
from plone.restapi.services import Service
from Products.CMFCore.permissions import ManagePortal
from typing import List
from zExceptions import BadRequest
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse

import json


DEFAULT_ID = "Plone"


def _is_outdated(site) -> bool:
    """Check if site needs an upgrade."""
    mig = getattr(site, "portal_migration", None) or site.get("portal_migration", None)
    if mig is not None:
        return mig.needUpgrading()
    return False


_no_content_marker = object()


@implementer(IPublishTraverse)
class SitesGet(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Consume any path segments after /@users as parameters
        self.params.append(name)
        return self

    def check_permission(self):
        """Ignore 'plone.restapi: Use REST API' permission on Application."""
        pass

    def render(self):
        # Make sure we do not change the order of the keys
        content = self.reply()
        if content is not _no_content_marker:
            self.request.response.setHeader("Content-Type", self.content_type)
            return json.dumps(content, indent=2, separators=(", ", ": "))

    def get_distributions(self, base_url: str) -> List[dict]:
        """Return a list of available distributions."""
        response = []
        distributions = dist_api.get_distributions()
        for dist in distributions:
            response.append(
                {
                    "@id": f"{base_url}/{dist.name}",
                    "name": dist.name,
                    "title": dist.title,
                    "description": dist.description,
                    "image": f"{self.context.absolute_url()}/@@dist-image/{dist.name}",
                }
            )
        return response

    def get_sites(self) -> List[dict]:
        """Return a list of site information."""
        response = []
        sites = site_api.get_sites(self.context)
        for site in sites:
            response.append(
                {
                    "@id": site.absolute_url(),
                    "id": site.id,
                    "title": site.title,
                    "description": site.description,
                    "creation_date": site.CreationDate(),
                    "needs_upgrade": _is_outdated(site),
                }
            )
        return response

    def _populate_server_defaults(self) -> dict:
        """Provide default values for new Plone sites."""
        request = self.request
        # Sites with default id
        all_sites = self.get_sites()
        sites = [site for site in all_sites if site["id"].startswith(DEFAULT_ID)]
        site_id = f"{DEFAULT_ID}{len(sites)}" if sites else DEFAULT_ID
        language = extract_browser_language(request)
        return {
            "site_id": site_id,
            "default_language": language,
        }

    def can_manage(self) -> bool:
        secman = getSecurityManager()
        return True if secman.checkPermission(ManagePortal, self.context) else False

    def reply(self) -> dict:
        base_url = f"{self.context.absolute_url()}/@sites"
        if len(self.params) == 0:
            return {
                "@id": base_url,
                "can_manage": self.can_manage(),
                "sites": self.get_sites(),
                "distributions": self.get_distributions(base_url),
            }
        else:
            name = self.params[0]
            dist = dist_api.get(name)
            if not dist:
                raise BadRequest("Parameters supplied are not valid")
            return {
                "@id": f"{base_url}/{dist.name}",
                "name": dist.name,
                "title": dist.title,
                "description": dist.description,
                "image": f"{base_url}/@@dist-image/{dist.name}",
                "schema": dist.schema,
                "uischema": dist.uischema,
                "default_values": self._populate_server_defaults(),
            }



"""Service to create a Plone Site."""
from plone.distribution.api import site as site_api
from plone.restapi.deserializer import json_body
from plone.restapi.services import Service
from zExceptions import BadRequest
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse

import plone.protect.interfaces


@implementer(IPublishTraverse)
class SiteCreate(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Consume any path segments after /@sites as parameters
        self.params.append(name)
        return self

    def reply(self):
        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)
        data = json_body(self.request)
        distribution_name = (
            self.params[0] if self.params else data.get("distribution", "default")
        )
        self.errors = []
        # self.validate_input_data(portal, data)
        # Create site
        try:
            site = site_api.create(
                self.context,
                distribution_name=distribution_name,
                answers=data,
            )
        except KeyError:
            raise BadRequest("Error creating the site.")
        else:
            return {
                "@id": site.absolute_url(),
                "id": site.id,
                "title": site.title,
                "description": site.description,
                "needs_upgrade": False,
            }



from plone.i18n.locales.interfaces import IContentLanguageAvailability
from zope.component import getGlobalSiteManager
from zope.component import getUtility
from zope.component import queryUtility
from zope.schema.interfaces import IVocabularyFactory


def _common_timezones():
    """Return a list of common timezones to be used during site creation."""
    context = getGlobalSiteManager()
    tz_vocab = getUtility(IVocabularyFactory, "plone.app.vocabularies.CommonTimezones")(
        context
    )

    response = []
    for term in tz_vocab:
        value = term.value
        splitted = value.split("/")
        friendly = splitted[-1]
        label = f"{friendly} ({value})"
        response.append({"label": label, "value": value})

    return response


def _available_languages() -> list:
    """Return available languages."""
    util = queryUtility(IContentLanguageAvailability)
    available = util.getLanguages(combined=True)
    languages = dict(util.getLanguageListing())

    # Group country specific versions by language
    grouped = {}
    for langcode, data in available.items():
        lang = langcode.split("-")[0]
        language = languages.get(lang, lang)  # Label

        struct = grouped.get(lang, {"label": language, "languages": []})

        langs = struct["languages"]
        langs.append(
            {
                "langcode": langcode,
                "label": data.get("native", data.get("name")),
            }
        )

        grouped[lang] = struct

    # Sort list by language, next by country
    response = []
    data = sorted(grouped.values(), key=lambda k: k["label"])
    for item in data:
        langs = sorted(item["languages"], key=lambda k: k["label"].lower())
        response.extend(langs)
    return response


def enrich_jsonschema(schema: dict) -> dict:
    """Process a jsonschema, adding definitions."""
    if "definitions" not in schema:
        schema["definitions"] = {}
    # Only add languages if does not exists
    if "languages" not in schema["definitions"]:
        languages = _available_languages()
        schema["definitions"]["languages"] = {
            "title": "Language",
            "type": "string",
            "default": "en",
            "description": "The main language of the site.",
            "anyOf": [
                {
                    "type": "string",
                    "enum": [lang["langcode"]],
                    "title": lang["label"],
                }
                for lang in languages
            ],
        }
    # Only add timezones if does not exists
    if "timezones" not in schema["definitions"]:
        # Set timezones
        schema["definitions"]["timezones"] = {
            "title": "Timezone",
            "type": "string",
            "default": "UTC",
            "description": (
                "The default timezone setting of the portal. "
                "Users will be able to set their own timezone, "
                "if available timezones are defined in the date "
                "and time settings."
            ),
            "anyOf": [
                {"type": "string", "enum": [tz["value"]], "title": tz["label"]}
                for tz in _common_timezones()
            ],
        }
    return schema


def enrich_uischema(uischema: dict, jsonschema: dict) -> dict:
    """Process a uischema and set default ordering (if not present)."""
    if "ui:order" not in uischema:
        # Set order to be the same as defined in the original schema creation.
        properties = [k for k in jsonschema["properties"].keys()]
        uischema["ui:order"] = properties
    return uischema


def process_raw_schema(raw_schema: dict) -> dict:
    """Process a schema file adding definitions and setting a default uischema."""
    jsonschema = raw_schema["schema"]
    uischema = enrich_uischema(raw_schema.get("uischema", {}), jsonschema)
    return {"schema": jsonschema, "uischema": uischema}


def validate_jsonschema(schema: dict, strict: bool = False) -> bool:
    """Validate if jsonschema has required information."""
    required_properties = ["site_id", "title", "setup_content"]
    recommended_properties = ["default_language", "portal_timezone"]
    properties = [key for key in schema.get("properties")]
    missing = [key for key in required_properties if key not in properties]
    if strict:
        missing = missing.extend(
            [key for key in recommended_properties if key not in properties]
        )
    return False if missing else True



from plone.distribution.core import Distribution
from Products.CMFPlone.Portal import PloneSite


def default_handler(
    distribution: Distribution, site: PloneSite, answers: dict
) -> PloneSite:
    """Default handler to create a new site."""
    # Process answers
    profiles = distribution.profiles
    setup_content = answers.get("setup_content", False)
    setup_tool = site["portal_setup"]
    for profile_id in profiles:
        setup_tool.runAllImportStepsFromProfile(f"profile-{profile_id}")

    # Add default content if needed
    if setup_content:
        contents = distribution.contents
        # First process any content profiles
        content_profiles = contents["profiles"]
        for profile_id in content_profiles:
            setup_tool.runAllImportStepsFromProfile(f"profile-{profile_id}")
        # Process content import from json
        content_json_path = contents["json"]
        if content_json_path:
            # TODO: Import content
            pass
    return site



from plone.app.testing import IntegrationTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer

import plone.distribution


class DistributionLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        self.loadZCML(package=plone.distribution)


FIXTURE = DistributionLayer()


INTEGRATION_TESTING = IntegrationTesting(
    bases=(FIXTURE,),
    name="DistributionLayer:IntegrationTesting",
)



from plone.distribution.api import distribution as dist_api
from zExceptions import NotFound
from zope.interface import implementer
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces import IPublishTraverse


@implementer(IPublishTraverse)
class DistributionImageView(BrowserView):
    def __init__(self, context, request):
        super().__init__(context, request)

    def publishTraverse(self, request, name):
        request["TraversalRequestNameStack"] = []
        self.path = name
        distribution_name = name.replace(".png", "")
        self.distribution = dist_api.get(distribution_name)
        if not self.distribution:
            raise NotFound(self.context)
        return self

    def set_headers(self, data, response):
        """Set response headers for the given file. If filename is given, set
        the Content-Disposition to attachment.
        """
        contenttype = "image/png"
        response.setHeader("Content-Type", contenttype)
        response.setHeader("Content-Length", len(data))

    def __call__(self):
        request = self.request
        distribution = self.distribution
        image = distribution.image
        data = image.read_bytes()
        self.set_headers(data, request.response)
        return data



from OFS.interfaces import IApplication
from plone.distribution.api import distribution as dist_api
from plone.distribution.api import site as site_api
from plone.distribution.core import Distribution
from plone.protect.interfaces import IDisableCSRFProtection
from Products.CMFPlone.browser.admin import AddPloneSite as AddPloneSiteView
from Products.CMFPlone.browser.admin import AppTraverser
from zope.component import adapts
from zope.component import queryMultiAdapter
from zope.interface import alsoProvides
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces import IRequest

import json


class RestTraverser(AppTraverser):
    """Traverser supporting REST calls in Zope Application root."""

    adapts(IApplication, IRequest)

    def publishTraverse(self, request, name):
        if not name.startswith("@"):
            return super().publishTraverse(request, name)

        service = queryMultiAdapter(
            (self.context, request), name=request._rest_service_id + name
        )
        if service is None:
            # No service, fallback to regular view
            view = queryMultiAdapter((self.context, request), name=name)
            if view is not None:
                return view
            raise
        return service


class Overview(BrowserView):
    """Overview page for Zope Root."""


class AddPloneSite(AddPloneSiteView):
    """Add Plone site."""

    distribution: Distribution = None
    default_data: dict = None

    def __call__(self):
        context = self.context
        form = self.request.form
        distribution_name = form.get("distribution", "default")
        site_id = form.get("site_id")
        self.distribution = dist_api.get(distribution_name)
        submitted = form.get("form.submitted", False)
        if submitted:
            # CSRF protect. DO NOT use auto CSRF protection for adding a site
            alsoProvides(self.request, IDisableCSRFProtection)
            site = site_api.create(
                context,
                distribution_name,
                profile_id="Products.CMFPlone:plone",
                answers=form,
            )
            self.request.response.redirect(site.absolute_url())
            return ""
        else:
            self.default_data = json.dumps(
                {
                    "site_id": site_id,
                    "default_language": self.browser_language(),
                    "portal_timezone": "UTC",
                },
                indent=2,
            )
        return self.index()



from AccessControl import getSecurityManager
from AccessControl.Permissions import view as View
from plone.base.interfaces import IPloneSiteRoot
from plone.distribution.api import distribution as dist_api
from plone.distribution.handler import default_handler
from plone.registry.interfaces import IRegistry
from Products.CMFPlone.events import SiteManagerCreatedEvent
from Products.CMFPlone.Portal import PloneSite
from Products.GenericSetup.tool import SetupTool
from typing import List
from ZODB.broken import Broken
from zope.component import queryUtility
from zope.component.hooks import setSite
from zope.event import notify
from zope.lifecycleevent import ObjectCreatedEvent


_TOOL_ID = "portal_setup"
_DEFAULT_PROFILE = "Products.CMFPlone:plone"


def _required_str_value(answers: dict, key: str) -> str:
    try:
        return answers[key]
    except KeyError:
        raise KeyError(f"A value for {key} is required.")


def get_sites(context=None) -> List[PloneSite]:
    """Get all Plone sites."""
    if not context:
        raise ValueError("Need to provide application root")
    result = []
    secman = getSecurityManager()
    candidates = (obj for obj in context.values() if not isinstance(obj, Broken))
    for obj in candidates:
        if obj.meta_type == "Folder":
            result.extend(get_sites(context=obj))
        elif IPloneSiteRoot.providedBy(obj):
            if secman.checkPermission(View, obj):
                result.append(obj)
        elif obj.getId() in getattr(context, "_mount_points", {}):
            result.extend(get_sites(context=obj))
    return result


def create(
    context,
    distribution_name: str,
    answers: dict,
    profile_id: str = _DEFAULT_PROFILE,
) -> PloneSite:
    """Create a new Plone site using one of the distributions."""
    distribution = dist_api.get(distribution_name)
    handler = distribution.handler if distribution.handler else default_handler
    post_handler = distribution.post_handler
    # Attributes used during site creation
    site_id = _required_str_value(answers, "site_id")
    title = _required_str_value(answers, "title")
    description = answers.get("description", "")
    default_language = answers.get("default_language", "en")
    portal_timezone = answers.get("portal_timezone", "UTC")
    # Create the Plone Site
    site = PloneSite(site_id)
    notify(ObjectCreatedEvent(site))
    context[site_id] = site
    site = context[site_id]
    # Set site language
    site.setLanguage(default_language)
    # Register SetupTool (portal_setup)
    site[_TOOL_ID] = SetupTool(_TOOL_ID)
    setup_tool = site[_TOOL_ID]
    notify(SiteManagerCreatedEvent(site))
    setSite(site)
    # Apply base profile
    setup_tool.setBaselineContext(f"profile-{profile_id}")
    setup_tool.runAllImportStepsFromProfile(f"profile-{profile_id}")
    # Set default properties (title, description)
    # Do this before applying extension profiles, so the settings from a
    # properties.xml file are applied and not overwritten by this
    props = dict(
        title=title,
        description=description,
    )
    site.manage_changeProperties(**props)
    # Set initial registry values
    reg = queryUtility(IRegistry, context=site)
    reg["plone.portal_timezone"] = portal_timezone
    reg["plone.available_timezones"] = [portal_timezone]
    reg["plone.default_language"] = default_language
    reg["plone.available_languages"] = [default_language]
    reg["plone.site_title"] = title
    # Run the Distribution handler
    site = handler(distribution, site, answers)
    # Run the Distribution post_handler
    if post_handler:
        site = post_handler(distribution, site, answers)
    return site



from plone.distribution.api import distribution as dist_api
from plone.distribution.core import Distribution

import pytest


class TestApiDistribution:
    def test_get_registry(self, integration):
        from plone.distribution.registry import DistributionRegistry

        registry = dist_api.get_registry()
        assert isinstance(registry, DistributionRegistry)

    def test_get_distributions(self, integration):
        distributions = dist_api.get_distributions()
        # In this package we have already 2 distributions
        assert len(distributions) >= 2
        # First distribution will be default
        dist = distributions[0]
        assert dist.name == "default"

    @pytest.mark.parametrize("name", ["default", "classic"])
    def test_get_success(self, integration, name):
        dist = dist_api.get(name=name)
        assert isinstance(dist, Distribution)
        assert dist.name == name

    @pytest.mark.parametrize("name", ["not_there", "fake"])
    def test_get_fail(self, integration, name):
        with pytest.raises(ValueError) as exc:
            dist_api.get(name=name)
        assert f"No distribution named {name}" in str(exc)



from zope.i18n.interfaces import IUserPreferredLanguages
from zope.i18n.locales import LoadLocaleError
from zope.i18n.locales import locales


def extract_browser_language(request) -> str:
    """Extract browser language from request."""
    language = "en"
    pl = IUserPreferredLanguages(request)
    if pl is not None:
        languages = pl.getPreferredLanguages()
        for httplang in languages:
            parts = (httplang.split("-") + [None, None])[:3]
            if parts[0] == parts[1]:
                # Avoid creating a country code for simple languages codes
                parts = [parts[0], None, None]
            try:
                locale = locales.getLocale(*parts)
                language = locale.getLocaleID().replace("_", "-").lower()
                break
            except LoadLocaleError:
                # Just try the next combination
                pass
    return language



from pathlib import Path
from plone.distribution import BASE_DISTRIBUTIONS_PATH
from plone.distribution.utils import schema as schema_utils
from typing import Callable
from typing import List

import json


DEFAULT_SCHEMA = json.load(open(BASE_DISTRIBUTIONS_PATH / "default" / "schema.json"))
DEFAULT_IMAGE = BASE_DISTRIBUTIONS_PATH / "default" / "image.png"


class Distribution:

    name: str
    title: str
    description: str
    directory: Path
    handler: Callable
    post_handler: Callable
    _schema: dict
    _profiles: dict

    def __init__(
        self,
        name: str,
        title: str,
        description: str,
        directory: Path,
        handler: Callable = None,
        post_handler: Callable = None,
    ):
        """Initialize a Plone Distribution."""
        self.name = name
        self.title = title
        self.description = description
        self.directory = directory
        self.handler = handler
        self.post_handler = post_handler
        schema_file = self.directory / "schema.json"
        raw_schema = DEFAULT_SCHEMA
        if schema_file.exists():
            try:
                raw_schema = json.loads(schema_file.read_text())
            except json.JSONDecodeError:
                # Log error
                pass
        schema = schema_utils.process_raw_schema(raw_schema)
        if not schema_utils.validate_jsonschema(schema["schema"]):
            raise ValueError(f"Invalid schema for {schema_file}")
        self._schema = schema
        profiles = {"base": [], "content": []}
        data_file = self.directory / "profiles.json"
        if data_file.exists():
            profiles = json.load(open(data_file))
        self._profiles = profiles

    def __repr__(self) -> str:
        return f"<Distribution name='{self.name}' title='{self.title}'>"

    @property
    def image(self) -> Path:
        """Return path to distribution image."""
        image = self.directory / "image.png"
        return image if image.exists() else DEFAULT_IMAGE

    @property
    def _content_folder(self) -> Path:
        """Return folder to example content."""
        folder = self.directory / "content"
        return folder

    @property
    def schema(self) -> dict:
        """Return a valid JSONSchema."""
        # We have to run enrich_jsonschema here because languages and
        # timezones may not be available during startup
        return schema_utils.enrich_jsonschema(self._schema["schema"])

    @property
    def uischema(self) -> dict:
        """Return the ui schema."""
        return self._schema["uischema"]

    @property
    def profiles(self) -> List[str]:
        """Return a list of profiles to be applied."""
        main_profiles = self._profiles.get("base", [])
        return main_profiles

    @property
    def contents(self) -> dict:
        """Return content structure."""
        content_folder = self._content_folder
        content_profiles = self._profiles.get("content", [])
        return {
            "profiles": content_profiles,
            "json": content_folder if content_folder.exists() else None,
        }



"""Authenticate a user in Zope root."""
from plone.restapi.deserializer import json_body
from plone.restapi.services import Service
from zope.interface import alsoProvides

import plone.protect.interfaces


class LoginPost(Service):
    def reply(self):
        """Authenticate on Zope Application root."""
        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)
        data = json_body(self.request)
        login = data.get("login")
        password = data.get("password")
        acl_users = self.context.acl_users
        if acl_users.authenticate(login, password, self.request):
            # sets cookie
            cookie_pas = acl_users.credentials_cookie_auth
            cookie_pas.updateCredentials(
                self.request, self.request.response, login, password
            )
            return {}
        else:
            self.request.response.setStatus(401)
            return {
                "error": {
                    "type": "Invalid credentials",
                    "message": "Wrong login and/or password.",
                }
            }



from pathlib import Path
from plone.distribution.core import Distribution
from plone.distribution.registry import _distribution_registry
from plone.distribution.registry import DistributionRegistry
from typing import List

import os


base_folder = Path(__file__).parent.parent / "distributions"


def _allow_list() -> List[str]:
    """Return a list of allowed distributions."""
    allowed_distributions = os.environ.get("ALLOWED_DISTRIBUTIONS", "")
    if allowed_distributions:
        allowed_distributions = allowed_distributions.split(",")
    return allowed_distributions


def get_registry() -> DistributionRegistry:
    """Return the Distribution Registry."""
    return _distribution_registry


def get_distributions(filter: bool = True) -> List[Distribution]:
    """Get available Plone distributions."""
    registry = get_registry()
    all_distributions = registry.enumerate_distributions()
    allowed_distributions = _allow_list()
    if filter and allowed_distributions:
        return [d for d in all_distributions if d.name in allowed_distributions]
    return all_distributions


def get(name: str) -> Distribution:
    """Get a distribution."""
    registry = get_registry()
    distribution = registry.lookup(name)
    if not distribution:
        raise ValueError(f"No distribution named {name}")
    return distribution



from pathlib import Path
from setuptools import find_packages
from setuptools import setup


long_description = f"""
{Path("README.md").read_text()}\n
{Path("CHANGES.md").read_text()}\n
"""

setup(
    name="plone.distribution",
    version="1.0.0a1",
    description="Plone distribution support",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Framework :: Zope :: 5",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.8",
    keywords="Plone CMF Python Zope CMS Distribution",
    author="Plone Foundation",
    author_email="releasemanager@plone.org",
    url="https://plone.org",
    license="GPL version 2",
    project_urls={
        "Homepage": "https://plone.org",
        "Documentation": "https://6.docs.plone.org",
        "Source": "https://github.com/plone/plone.distribution",
        "Issues": "https://github.com/plone/plone.distribution/issues",
    },
    packages=find_packages("src"),
    namespace_packages=["plone"],
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "Plone",
        "setuptools",
        "z3c.unconfigure",
    ],
    extras_require={
        "test": [
            "zest.releaser[recommended]",
            "zestreleaser.towncrier",
            "plone.app.testing",
            "plone.restapi[test]",
            "pytest",
            "pytest-cov",
            "pytest-plone>=0.2.0",
        ]
    },
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    """,
)



from plone.distribution.core import Distribution
from plone.distribution.registry import _distribution_registry
from types import FunctionType
from zope.configuration.fields import GlobalObject
from zope.configuration.fields import MessageID
from zope.configuration.fields import Path
from zope.configuration.name import resolve
from zope.interface import Interface

import pathlib
import zope.schema


def _resolveDottedName(dotted):
    __traceback_info__ = dotted

    try:
        return resolve(dotted)
    except ModuleNotFoundError:
        return


class IRegisterDistributionDirective(Interface):
    """Register distributions with the global registry."""

    name = zope.schema.TextLine(
        title="Name", description="Distribution short name.", required=True
    )

    title = MessageID(
        title="Title", description="Distribution Title.", default="", required=True
    )

    description = MessageID(
        title="Description",
        description="Optional description for the distribution.",
        default="",
        required=False,
    )

    handler = GlobalObject(
        title="Handler",
        description=("Function called after initial site creation. "),
        required=False,
    )

    post_handler = GlobalObject(
        title="Post Handler",
        description=(
            "Function called after site is created and handler processed profiles."
        ),
        required=False,
    )

    directory = Path(
        title="Distribution Directory",
        description="If not specified 'distributions/<name>' is used.",
        required=False,
    )


def _check_function(name: str, product: str, func: FunctionType = None) -> FunctionType:
    if func and not isinstance(func, FunctionType):
        # Probably a string, so resolve dotted name
        func = f"{product}{func}" if func.startswith(".") else func
        func = _resolveDottedName(func)
        if not func:
            raise ValueError(f"{name} points to non existing function: {func}")
    return func


def register_distribution(
    _context,
    name,
    title,
    description="",
    directory=None,
    handler=None,
    post_handler=None,
):
    """Add a new distribution to the registry."""
    product = _context.package.__name__
    package_path = pathlib.Path(_context.package.__file__).parent.resolve()
    if directory is None:
        directory = f"distributions/{name}"
    directory = package_path / directory

    if description is None:
        description = ""

    # Validate handler and post_handler functions
    handler = _check_function("handler", product, handler)
    post_handler = _check_function("post_handler", product, post_handler)

    distribution = Distribution(
        name, title, description, directory, handler, post_handler
    )

    _context.action(
        discriminator=("register", product, name),
        callable=_distribution_registry.register,
        args=(distribution,),
    )



from zope.interface import Interface


class IDistributionRegistry(Interface):
    def register(distribution):
        """Register a distribution in the Distribution Registry."""

    def unregister(distribution):
        """Remove registration of a distribution in the Distribution Registry."""

    def clear():
        """Remove all distribution's registrations."""

    def lookup(name):
        """Return a named distribution."""

    def __getitem__(name):
        """Convenience spelling for `lookup(name)`."""

    def enumerate_distributions():
        """Return all distributions registered in the Distribution Registry."""


class IDistribution(Interface):
    """Named distribution."""



from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from Acquisition import Implicit
from plone.distribution.core import Distribution
from plone.distribution.interfaces import IDistribution
from plone.distribution.interfaces import IDistributionRegistry
from typing import Any
from typing import List
from typing import Tuple
from zope.component import getGlobalSiteManager
from zope.interface import implementer
from zope.interface import Interface


ManagePortal = "Manage Site"


class GlobalRegistryStorage:
    """Helper to manage Registry access."""

    def __init__(self, interfaceClass: Interface):
        self.interfaceClass = interfaceClass

    def keys(self) -> List[str]:
        """Return a list of names of all registered utilities."""
        sm = getGlobalSiteManager()
        return [n for n, _i in sm.getUtilitiesFor(self.interfaceClass)]

    def values(self) -> List[Any]:
        """Return a list with all registered utilities."""
        sm = getGlobalSiteManager()
        return [i for _n, i in sm.getUtilitiesFor(self.interfaceClass)]

    def items(self) -> List[Tuple[str, Any]]:
        """Return a list with all registered utilities as tuples.

        Each tuple contains name, utility.
        """
        sm = getGlobalSiteManager()
        return [(n, i) for n, i in sm.getUtilitiesFor(self.interfaceClass)]

    def get(self, key: str) -> Any:
        """Return a registered utility by name."""
        sm = getGlobalSiteManager()
        return sm.queryUtility(provided=self.interfaceClass, name=key)

    def __setitem__(self, id: str, info: Any) -> None:
        """Register a named utility."""
        sm = getGlobalSiteManager()
        return sm.registerUtility(info, provided=self.interfaceClass, name=id)

    def __delitem__(self, id: str) -> bool:
        """Remove a registration for a named utility."""
        sm = getGlobalSiteManager()
        return sm.unregisterUtility(provided=self.interfaceClass, name=id)

    def clear(self) -> None:
        """Remove all registration of named utilities."""
        for key in self.keys():
            del self[key]


@implementer(IDistributionRegistry)
class DistributionRegistry(Implicit):
    """Track Registered Distributions."""

    security = ClassSecurityInfo()
    security.setDefaultAccess("allow")

    def __init__(self):
        """Initialize the Distribution Registration"""
        self._registered = GlobalRegistryStorage(IDistribution)
        self.clear()

    @security.protected(ManagePortal)
    def register(self, distribution: Distribution):
        """Register a distribution in the Distribution Registry."""
        self._registered[distribution.name] = distribution

    @security.protected(ManagePortal)
    def unregister(self, distribution: Distribution):
        """Remove registration of a distribution in the Distribution Registry."""
        del self._registered[distribution.name]

    @security.private
    def clear(self):
        """Remove all distribution's registrations."""
        self._registered.clear()

    @security.protected(ManagePortal)
    def lookup(self, name: str) -> Distribution:
        """Return a named distribution."""
        return self._registered.get(name)

    @security.protected(ManagePortal)
    def __getitem__(self, name) -> Distribution:
        """Convenience spelling for `lookup(name)`."""
        return self.lookup(name)

    @security.protected(ManagePortal)
    def enumerate_distributions(self) -> List[Distribution]:
        """Return all distributions registered in the Distribution Registry.

        This implementation returns the distributions sorted by name, but keep
        the distribuion named default as the first item of the list.
        """
        result = []
        distributions = sorted(self._registered.items())
        for name, distribution in distributions:
            if name == "default":
                result.insert(0, distribution)
            else:
                result.append(distribution)
        return tuple(result)


InitializeClass(DistributionRegistry)

_distribution_registry = DistributionRegistry()
